from __future__ import annotations

from typing import TYPE_CHECKING

from spotdl import Spotdl
from spotdl.types.options import DownloaderOptions
from spotdl.utils.config import DOWNLOADER_OPTIONS

from music.models import Path

if TYPE_CHECKING:
    from music.context import Secrets  # pragma: nocover


class Client(Spotdl):  # type: ignore[misc]
    @classmethod
    def create(cls, secrets: Secrets) -> Client:
        options = cls.create_downloader_options()
        return cls(
            secrets.spotify.client_id,
            secrets.spotify.client_secret,
            downloader_settings=options,
        )

    @classmethod
    def create_downloader_options(cls) -> DownloaderOptions:
        output_format = str(Path.downloaded_songs / "{artists} - {title}.{output-ext}")
        custom_options = {"output": output_format, "format": "opus", "threads": 10}
        options = DOWNLOADER_OPTIONS | custom_options
        return DownloaderOptions(**options)



import cli

from music.models import Path
from music.runtime import runtime
from music.utils.batched import batched

from . import spotdl
from .downloader import Downloader


def download_new_songs() -> None:
    fix_song_count()
    size = runtime.context.config.download_chunk_size
    ids_to_download = runtime.storage.ids_to_download
    for ids in batched(ids_to_download, size=size):
        Downloader(ids).start()
    download_youtube_songs()


def download_youtube_songs() -> None:
    urls = [
        f"https://www.youtube.com/watch?v={id_}"
        for id_ in runtime.storage.youtube_tracks_to_download
    ]
    if urls:
        options = {
            "output": str(Path.processed_songs / "%(title)s.%(ext)s"),
            "embed-thumbnail": None,
            "embed-metadata": None,
            "audio-format": "opus",
        }
        cli.run("yt-dlp", options, "-x", urls)
        runtime.storage.youtube_tracks_to_download = []


def fix_song_count() -> None:
    num_songs = len(runtime.storage.tracks_to_download)
    progress_handler = spotdl.downloader.progress_handler
    progress_handler.set_song_count(num_songs)
    progress_handler.set_song_count = lambda _: None



from dataclasses import dataclass
from functools import cached_property
from typing import Any

from package_utils.dataclasses.mixins import SerializationMixin

dataclass = dataclass(eq=False)  # type: ignore[assignment]

POPULARITY_THRESHOLD = 15
MIN_SONG_DURATION = 2 * 60 * 1000
MAX_SONG_DURATION = 10 * 60 * 1000


@dataclass
class Urls:
    spotify: str


@dataclass
class IDS:
    isrc: str | None


@dataclass
class Item(SerializationMixin):
    external_urls: Urls
    href: str
    id: str
    name: str
    type: str
    uri: str

    def __hash__(self) -> int:
        return hash(self.id)


@dataclass
class Followers:
    href: str | None
    total: int


@dataclass
class Image:
    height: int
    url: str
    width: int


@dataclass
class Artist(Item):
    pass


@dataclass
class ArtistInfo(Item):
    followers: Followers
    genres: list[str]
    images: list[Image]
    popularity: int


@dataclass
class Album(Item):
    album_type: str
    artists: list[Artist]
    images: list[Image]
    release_date: str
    release_date_precision: str
    total_tracks: int
    album_group: str | None
    is_playable: bool | None
    available_markets: list[str] | None
    restrictions: dict[str, str] | None


@dataclass
class AlbumTrack(Item):
    artists: list[Artist]
    disc_number: int
    duration_ms: int
    explicit: bool
    is_local: bool
    is_playable: bool | None
    preview_url: str | None
    track_number: int
    linked_from: dict[str, Any] | None
    restrictions: dict[str, str] | None


@dataclass
class Track(AlbumTrack):
    album: Album
    popularity: int
    external_ids: IDS
    available_markets: list[str] | None

    @property
    def should_download(self) -> bool:
        skip_names = ("Interlude", "Intro", "Outro", "Live", "Instrumental")
        return (
            MIN_SONG_DURATION < self.duration_ms < MAX_SONG_DURATION
            and self.popularity > POPULARITY_THRESHOLD
            and not any(f" - {skip_name}" in self.name for skip_name in skip_names)
        )

    @cached_property
    def full_name(self) -> str:
        artist_names = ", ".join(artist.name for artist in self.artists)
        return f"{artist_names} - {self.name}"


@dataclass
class PaginatedResponse(SerializationMixin):
    href: str
    previous: str | None
    next: str | None
    limit: int
    offset: int
    total: int


@dataclass
class ArtistSearch(PaginatedResponse):
    items: list[ArtistInfo]


@dataclass
class Artists(SerializationMixin):
    artists: list[ArtistInfo]


@dataclass
class Tracks(SerializationMixin):
    tracks: list[Track]


@dataclass
class Albums(PaginatedResponse):
    items: list[Album]


@dataclass
class AlbumTracks(PaginatedResponse):
    items: list[AlbumTrack]


@dataclass
class TrackSearch(PaginatedResponse):
    items: list[Track]


@dataclass
class RecommendedTracks(Tracks):
    seeds: list[dict[str, str | int]]



import random
from collections.abc import Iterator
from itertools import islice
from typing import Any

from music.models.response_types import ArtistInfo
from music.runtime import runtime


class Server:
    @classmethod
    def create_artist_search_results(cls, name: str) -> Iterator[dict[str, Any]]:
        saved_artist_ids = runtime.storage.artist_ids
        artists = runtime.spotify_client.search_artist(name)
        for artist in artists:
            is_added = artist.id in saved_artist_ids
            yield artist.dict() | {"added": is_added}

    @classmethod
    def create_artist_recommendations(cls) -> list[ArtistInfo]:
        recommendations = cls._create_artist_recommendations()
        cls.increase_recommendation_frequencies(recommendations)
        frequencies = runtime.storage.recommendation_frequencies
        return sorted(recommendations, key=lambda artist: frequencies[artist.id])

    @classmethod
    def _create_artist_recommendations(cls) -> set[ArtistInfo]:
        iterator = cls.generate_recommended_artists()
        limited_iterator = islice(
            iterator,
            runtime.context.config.number_of_recommendations,
        )
        return set(limited_iterator)

    @classmethod
    def generate_recommended_artists(cls) -> Iterator[ArtistInfo]:
        saved_artist_ids = list(runtime.storage.artist_ids)
        random.shuffle(saved_artist_ids)
        seed_ids = saved_artist_ids[
            : runtime.context.config.max_new_recommendation_tries
        ]
        for id_ in seed_ids:
            related_artists = runtime.spotify_client.related_artists(id_)
            for artist in related_artists:
                if artist.id not in saved_artist_ids:
                    yield artist

    @classmethod
    def increase_recommendation_frequencies(cls, artists: set[ArtistInfo]) -> None:
        frequencies = runtime.storage.recommendation_frequencies
        for artist in artists:
            frequencies[artist.id] = frequencies.get(artist.id, 0) + 1
        runtime.storage.recommendation_frequencies = frequencies

    @classmethod
    def load_saved_artists(
        cls,
        offset: int,
        limit: int | None,
    ) -> Iterator[dict[str, str]]:
        artists = runtime.storage.artists
        limited_artists = (
            artists[offset:] if limit is None else artists[offset : offset + limit]
        )
        ids = [artist.id for artist in limited_artists]
        api_infos = runtime.spotify_client.artists(ids)
        for artist, info in zip(limited_artists, api_infos, strict=False):
            yield info.dict() | artist.dict()



from package_utils.context.entry_point import create_entry_point
from webapp_starter.context.context import Context_, context
from webapp_starter.main.main import main
from webapp_starter.models import Config, Options, Secrets


def add_name(created_context: Context_[Options, Config, Secrets]) -> None:
    created_context.options.name = "music"


entry_point = create_entry_point(main, context, context_creation_callback=add_name)
